package com.css.simulation.resource.scheduler.common.util;

import api.common.pojo.constants.DictConstants;
import api.common.pojo.dto.ProjectMessageDTO;
import api.common.util.*;
import com.css.simulation.resource.scheduler.common.configuration.kubernetes.KubernetesConfiguration;
import com.css.simulation.resource.scheduler.dao.entity.ProjectPO;
import com.css.simulation.resource.scheduler.dao.entity.UserPO;
import com.css.simulation.resource.scheduler.dao.mapper.AutoSubProjectMapper;
import com.css.simulation.resource.scheduler.dao.mapper.ClusterMapper;
import com.css.simulation.resource.scheduler.dao.mapper.ManualProjectMapper;
import com.css.simulation.resource.scheduler.dao.mapper.UserMapper;
import com.css.simulation.resource.scheduler.service.domain.KubernetesNodeTO;
import com.css.simulation.resource.scheduler.service.domain.NodeTO;
import com.css.simulation.resource.scheduler.service.domain.PrefixTO;
import io.kubernetes.client.openapi.ApiClient;
import lombok.SneakyThrows;
import lombok.Synchronized;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.kafka.core.KafkaTemplate;
import org.springframework.stereotype.Component;

import javax.annotation.Resource;
import java.io.File;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * 任务结果打分
 */
@Component
@Slf4j
public class ProjectUtil {

    @Value("${scheduler.linux-path.pod-yaml-directory}")
    private String podYamlDirectory;
    @Value("${scheduler.host.hostname}")
    private String hostname;
    @Value("${scheduler.host.username}")
    private String username;
    @Value("${scheduler.host.password}")
    private String password;


    @Resource
    private ManualProjectMapper manualProjectMapper;
    @Resource
    private AutoSubProjectMapper autoSubProjectMapper;
    @Resource
    private UserMapper userMapper;
    @Resource
    private ClusterMapper clusterMapper;
    @Resource
    private KafkaTemplate<String, String> kafkaTemplate;
    @Resource
    private KubernetesConfiguration kubernetesConfiguration;
    @Resource
    private StringRedisTemplate stringRedisTemplate;
    @Resource
    private ApiClient apiClient;


    @SneakyThrows
    public void deleteYamlByProjectId(String projectId) {
        List<String> absolutePathList = FileUtil.listAbsolutePath(podYamlDirectory);
        for (String absolutePath : absolutePathList) {
            if (absolutePath.contains(projectId)) {
                boolean delete = new File(absolutePath).delete();
            }
        }
    }


    public List<NodeTO> getNodeListToCount(Map<String, Integer> nodeMap) {
        List<NodeTO> result = new ArrayList<>();
        nodeMap.forEach((nodeName, parallelism) -> {
            for (int i = 0; i < parallelism; i++) {
                result.add(new NodeTO(nodeName, 0));
            }
        });
        return result;
    }

    /**
     * 判断算法是否已经导入
     *
     * @return 算法是否已经导入
     */
    @SneakyThrows
    public boolean isImported(String dockerImageWithoutVersion) {
        String dockerImageListString = LinuxUtil.execute("docker images");
        return dockerImageListString.contains(dockerImageWithoutVersion);
    }

    public String getRandomPodName(String projectId) {
        return "project-" + projectId + "-" + StringUtil.getRandomEightBitUUID();
    }

    public String getPodYamlName(String nodeName, String podName) {
        return nodeName + "#" + podName + ".yaml";
    }


    public void deletePod(String podName) {
        try {
            String key = "pod:" + podName + ":node";
            // 先删除 redis key
            RedisUtil.deleteByKey(stringRedisTemplate, "pod:" + podName + ":node");
            KubernetesUtil.deletePod(apiClient, kubernetesConfiguration.getNamespace(), podName);
            TimeUnit.SECONDS.sleep(10); // 暂停 10 秒等待资源完全释放。
            log.info("ProjectUtil--deletePod 删除 pod 并删除 redis 键值对：" + key);
        } catch (Exception e) {
            e.printStackTrace();
            log.error("ProjectUtil--deletePod pod " + podName + " 不存在。", e);
        }
    }


    public String getNodeNameOfPod(String podName) {
        return stringRedisTemplate.opsForValue().get("pod:" + podName + ":node");
    }

    /**
     * 更改一个名字继续启动
     *
     * @param projectId   项目 id
     * @param nodeName    运行 pod 的节点名称
     * @param lastPodName 即将删除的 pod 名称
     */
    @SneakyThrows
    public void createNextPod2(String projectId, String nodeName, String lastPodName) {
        //1 删除上一个 pod 和 redis 键值对 和 旧的 yaml 文件
//        new Thread(() -> deletePod(lastPodName), "delete-" + lastPodName).start();
        //
        deletePod(lastPodName);
        List<String> list = FileUtil.listAbsolutePath(podYamlDirectory);
        Iterator<String> iterator1 = list.iterator();
        while (iterator1.hasNext()) {
            String absolutePath = iterator1.next();
            if (absolutePath.contains(nodeName) && absolutePath.contains(lastPodName)) {
                FileUtil.rm(absolutePath);
                list.remove(absolutePath);
                break;
            }
        }
        for (String absolutePath : list) {
            if (absolutePath.contains(nodeName) && absolutePath.contains(projectId)) {
                String cpuOrderString = stringRedisTemplate.opsForValue().get("pod:" + lastPodName + ":cpu");
                Optional.ofNullable(cpuOrderString).orElseThrow(() -> new RuntimeException("createNextPod2() pod " + lastPodName + " 缓存的 cpu 编号为空。"));
                final String read = FileUtil.read(absolutePath);
                final String replace = read.replace("cpu-order", "\"" + cpuOrderString + "\"");
                FileUtil.writeStringToLocalFile(replace, absolutePath);
                createPod2(absolutePath);
                log.info("createNextPod2() 创建项目 " + projectId + " 在节点 " + nodeName + " 的下一个 pod。");
                return;
            }
        }
        // 如果当前节点没有下一个yaml，则返回一个并行度。
        log.info("createNextPod2() 节点 " + nodeName + " 已经执行完被分配的项目 " + projectId + " 的所有 pod。");
        addOneParallelismToNode(nodeName);
    }


    /**
     * @param nodeName       节点名称
     * @param podName        pod 名称
     * @param podYamlContent pod 文件内容
     */
    @SneakyThrows
    public void createPod(String nodeName, String podName, String podYamlContent) {
        stringRedisTemplate.opsForValue().set("pod:" + podName + ":node", nodeName);    // 将 pod 运行在哪个 node 上记录到 redis
        String podYamlName = getPodYamlName(nodeName, podName);
        String podYamlPath = podYamlDirectory + podYamlName;
        FileUtil.writeStringToLocalFile(podYamlContent, podYamlPath);
        KubernetesUtil.createNs(apiClient, kubernetesConfiguration.getNamespace());
        KubernetesUtil.applyYaml(hostname, username, password, podYamlPath);
//        下面这种执行方法报错
//        V1Pod v1Pod = (V1Pod) Yaml.load(nextPodString);
//        KubernetesUtil.createPod(apiClient, kubernetesNamespace, v1Pod);
    }

    /**
     * @param podYamlPath pod 文件内容
     */
    public void createPod2(String podYamlPath) {
        String nodeName = new File(podYamlPath).getName().split("#")[0];
        String podName = podYamlPath.split("#")[1].split("\\.")[0];
        stringRedisTemplate.opsForValue().set("pod:" + podName + ":node", nodeName);    // 将 pod 运行在哪个 node 上记录到 redis
        new Thread(() -> KubernetesUtil.applyYaml(hostname, username, password, podYamlPath), "create-" + podName).start();
    }


    public String getProjectTypeByProjectId(String projectId) {
        String projectType = null;
        ProjectPO manualProjectPO = manualProjectMapper.selectById(projectId);
        ProjectPO autoSubProjectPO = autoSubProjectMapper.selectById(projectId);
        if (manualProjectPO != null) {
            projectType = DictConstants.PROJECT_TYPE_MANUAL;
        } else if (autoSubProjectPO != null) {
            projectType = DictConstants.PROJECT_TYPE_AUTO_SUB;
        }
        return projectType;
    }

    public ProjectPO getProjectByProjectId(String projectId) {
        ProjectPO manualProjectPO = manualProjectMapper.selectById(projectId);
        ProjectPO autoSubProjectPO = autoSubProjectMapper.selectById(projectId);
        if (manualProjectPO != null) {
            return manualProjectPO;
        } else if (autoSubProjectPO != null) {
            return autoSubProjectPO;
        }
        throw new RuntimeException("ProjectUtil--getProjectByProjectId 不存在项目：" + projectId);
    }


    /**
     * 获取正在运行的项目的并行度总和
     *
     * @param clusterRunningPrefix 集群 key 前缀
     * @return 正在运行的项目的并行度总和
     */
    @SneakyThrows
    public int getCurrentParallelismSum(String clusterRunningPrefix) {
        int result = 0;
        Set<String> clusterRunningKeySet = stringRedisTemplate.keys(clusterRunningPrefix + "*");
        List<String> runningProjectSet; // 运行中的 projectId 列表
        if (CollectionUtil.isEmpty(clusterRunningKeySet)) {
            return 0;
        }
        runningProjectSet = getRunningProjectList(clusterRunningKeySet);
        if (CollectionUtil.isEmpty(runningProjectSet)) {
            return 0;
        }
        for (String projectKey : runningProjectSet) {
            String projectJsonTemp = stringRedisTemplate.opsForValue().get(projectKey);
            ProjectMessageDTO projectMessageTemp = JsonUtil.jsonToBean(projectJsonTemp, ProjectMessageDTO.class);
            result += projectMessageTemp.getCurrentParallelism();   // 获取当前正在使用的并行度
        }
        return result;
    }


    /**
     * 节点剩余可用并行度列表
     *
     * @return 节点映射（节点名，并行度）
     */
    public Map<String, Integer> getNodeMap() {
        List<KubernetesNodeTO> initialNodeList = kubernetesConfiguration.getNodeList(); // 预设并行度的节点列表
        log.info("ProjectUtil--getNodeMap 预设并行度的节点列表为：" + initialNodeList);
        Map<String, Integer> resultNodeMap = new HashMap<>();    // 用于执行的节点映射（节点名，并行度）
        for (KubernetesNodeTO kubernetesNodeSource : initialNodeList) {
            KubernetesNodeTO kubernetesNodeCopy = kubernetesNodeSource.clone();
            String nodeName = kubernetesNodeCopy.getName();
            int maxParallelism = kubernetesNodeCopy.getMaxParallelism();
            String restParallelismKey = "node:" + nodeName + ":parallelism";
            String restParallelismString = stringRedisTemplate.opsForValue().get(restParallelismKey);
            int restParallelism;
            if (restParallelismString == null) {    // 如果剩余可用并行度没有值，说明是第一次查询，则重置成最大并行度的预设值
                restParallelism = maxParallelism;
                stringRedisTemplate.opsForValue().set(restParallelismKey, restParallelism + "");
            } else {
                restParallelism = Integer.parseInt(restParallelismString);
            }
            resultNodeMap.put(nodeName, restParallelism);
        }
        log.info("getNodeMap() 剩余并行度的节点列表为：" + resultNodeMap);
        return resultNodeMap;
    }


    /**
     * 获取集群剩余并行度
     *
     * @return 集群剩余并行度
     */
    public int getRestParallelism() {
        List<KubernetesNodeTO> initialNodeList = kubernetesConfiguration.getNodeList(); // 预设并行度的节点列表
        // 遍历所有节点，获取还有剩余并行度的节点
        List<KubernetesNodeTO> restNodeList = new ArrayList<>();    // 剩余并行度的节点列表
        for (KubernetesNodeTO kubernetesNodeSource : initialNodeList) {
            KubernetesNodeTO kubernetesNodeCopy = kubernetesNodeSource.clone();
            String nodeName = kubernetesNodeCopy.getName();   // 节点名称
            int maxParallelism = kubernetesNodeCopy.getMaxParallelism();
            String restParallelismString = stringRedisTemplate.opsForValue().get("node:" + nodeName + ":parallelism");// 获取节点剩余并行度的 key
            // -------------------------------- Comment --------------------------------
            int restParallelism;
            if (restParallelismString == null || Integer.parseInt(restParallelismString) > maxParallelism) {    // 如果剩余可用并行度没有值，说明是第一次查询，则重置成最大并行度的预设值
                restParallelism = maxParallelism;
                stringRedisTemplate.opsForValue().set("node:" + nodeName + ":parallelism", restParallelism + "");
            } else {
                restParallelism = Integer.parseInt(restParallelismString);
                kubernetesNodeCopy.setMaxParallelism(restParallelism);
            }
            if (restParallelism > 0) {
                restNodeList.add(kubernetesNodeCopy);
            }
        }
        log.info("ProjectUtil--getRestParallelism 集群剩余并行度为：" + restNodeList);
        return restNodeList.size() == 0 ? 0 : restNodeList.stream().mapToInt(KubernetesNodeTO::getMaxParallelism).sum();
    }

    /**
     * 根据并行度获取用于执行的节点列表
     * 根据剩余可用并行度降序排序
     *
     * @return 节点映射（节点名，并行度）
     */
    public Map<String, Integer> getNodeMapToUse(int parallelism) {
        List<KubernetesNodeTO> initialNodeList = kubernetesConfiguration.getNodeList(); // 预设并行度的节点列表
        log.info("ProjectUtil--getNodeMapToUse 预设并行度的节点列表为：" + initialNodeList);
        // 遍历所有节点，获取还有剩余并行度的节点
        List<KubernetesNodeTO> restNodeList = new ArrayList<>();    // 剩余并行度的节点列表
        for (KubernetesNodeTO kubernetesNodeSource : initialNodeList) {
            KubernetesNodeTO kubernetesNodeCopy = kubernetesNodeSource.clone();
            String nodeName = kubernetesNodeCopy.getName();   // 节点名称
            int maxParallelism = kubernetesNodeCopy.getMaxParallelism();
            String restParallelismString = stringRedisTemplate.opsForValue().get("node:" + nodeName + ":parallelism");// 获取节点剩余并行度的 key
            // -------------------------------- Comment --------------------------------
            int restParallelism;
            if (restParallelismString == null || Integer.parseInt(restParallelismString) > maxParallelism) {    // 如果剩余可用并行度没有值，说明是第一次查询，则重置成最大并行度的预设值
                restParallelism = maxParallelism;
                stringRedisTemplate.opsForValue().set("node:" + nodeName + ":parallelism", restParallelism + "");
            } else {
                restParallelism = Integer.parseInt(restParallelismString);
                kubernetesNodeCopy.setMaxParallelism(restParallelism);
            }
            if (restParallelism > 0) {
                restNodeList.add(kubernetesNodeCopy);
            }
        }
        log.info("ProjectUtil--getNodeMapToUse 剩余并行度的节点列表为：" + restNodeList);
        Map<String, Integer> resultNodeMap = new HashMap<>();    // 用于执行的节点映射（节点名，并行度）
        if (!CollectionUtil.isEmpty(restNodeList)) {
            if (restNodeList.size() == 1) {
                KubernetesNodeTO tempNode = restNodeList.get(0);
                String tempNodeName = tempNode.getName();
                int tempParallelism = tempNode.getMaxParallelism();
                resultNodeMap.put(tempNodeName, Math.min(tempParallelism, parallelism));
            }
            if (restNodeList.size() > 1) {
                for (int i = 0; i < parallelism; i++) {
                    // 每次降序排序都取剩余并行度最大的一个。
                    restNodeList.sort((o1, o2) -> o2.getMaxParallelism() - o1.getMaxParallelism());
                    KubernetesNodeTO tempNode = restNodeList.get(0);
                    String tempNodeName = tempNode.getName();
                    int tempParallelism = tempNode.getMaxParallelism();
                    if (tempParallelism > 0) {
                        tempNode.setMaxParallelism(tempParallelism - 1);
                        CollectionUtil.addValueToMap(resultNodeMap, 1, tempNodeName);
                    }
                }
            }
        }
        log.info("ProjectUtil--getNodeMapToUse 即将使用节点的并行度为：" + resultNodeMap);
        return resultNodeMap;
    }


    public PrefixTO getRedisPrefixByUserIdAndProjectIdAndTaskId(String userId, String projectId, String taskId) {
        //3 获取用户类型（管理员账户、管理员子账户、普通账户、普通子账户）（独占、共享）
        UserPO userPO = userMapper.selectById(userId);
        String roleCode = userPO.getRoleCode();
        String useType = userPO.getUseType();
        String clusterId;
        if (DictConstants.ROLE_CODE_SYSADMIN.equals(roleCode) || DictConstants.ROLE_CODE_ADMIN.equals(roleCode)) {  //3-1 管理员账户和管理员子账户直接执行
            clusterId = DictConstants.SYSTEM_CLUSTER_ID;
        } else if (DictConstants.ROLE_CODE_UESR.equals(roleCode)) { //3-2 普通账户，不管是独占还是共享，都在自己的集群里排队，根据自己的独占节点排队
            clusterId = clusterMapper.selectByUserId(userId).getId();
        } else if (DictConstants.ROLE_CODE_SUBUESR.equals(roleCode)) {
            if (DictConstants.USER_TYPE_EXCLUSIVE.equals(useType)) {   //3-3 普通子账户，根据自己的独占节点排队
                clusterId = clusterMapper.selectByUserId(userId).getId();
            } else {    //3-4 共享子账户，根据父账户的共享节点排队
                String parentUserId = userPO.getCreateUserId();
                clusterId = clusterMapper.selectByUserId(parentUserId).getId();
            }
        } else {
            throw new RuntimeException("ProjectUtil--getRedisPrefixByUserIdAndProjectIdAndTaksId 未知账户类型，无法获取集群信息！");
        }
        String clusterPrefix = "cluster:" + clusterId;
        String clusterRunningPrefix = clusterPrefix + ":running";
        String projectRunningKey = clusterRunningPrefix + ":" + projectId;
        String taskTickKey = projectRunningKey + ":task:" + taskId + ":tick";
        String taskPodKey = projectRunningKey + ":task:" + taskId + ":pod";
        String taskRetryKey = projectRunningKey + ":task:" + taskId + ":retry";
        String taskMessageKey = projectRunningKey + ":task:" + taskId + ":message";

        return PrefixTO.builder()
                .clusterPrefix(clusterPrefix)
                .clusterRunningPrefix(clusterRunningPrefix)
                .projectRunningKey(projectRunningKey)
                .taskTickKey(taskTickKey)
                .taskPodKey(taskPodKey)
                .taskRetryKey(taskRetryKey)
                .taskMessageKey(taskMessageKey)
                .build();

    }


    public PrefixTO getRedisPrefixByUserIdAndProjectId(String userId, String projectId) {
        //3 获取用户类型（管理员账户、管理员子账户、普通账户、普通子账户）（独占、共享）
        UserPO userPO = userMapper.selectById(userId);
        String roleCode = userPO.getRoleCode();
        String useType = userPO.getUseType();
        String clusterId;
        if (DictConstants.ROLE_CODE_SYSADMIN.equals(roleCode) || DictConstants.ROLE_CODE_ADMIN.equals(roleCode)) {  //3-1 管理员账户和管理员子账户直接执行
            clusterId = DictConstants.SYSTEM_CLUSTER_ID;
        } else if (DictConstants.ROLE_CODE_UESR.equals(roleCode)) { //3-2 普通账户，不管是独占还是共享，都在自己的集群里排队，根据自己的独占节点排队
            clusterId = clusterMapper.selectByUserId(userId).getId();
        } else if (DictConstants.ROLE_CODE_SUBUESR.equals(roleCode)) {
            if (DictConstants.USER_TYPE_EXCLUSIVE.equals(useType)) {   //3-3 普通子账户，根据自己的独占节点排队
                clusterId = clusterMapper.selectByUserId(userId).getId();
            } else {    //3-4 共享子账户，根据父账户的共享节点排队
                String parentUserId = userPO.getCreateUserId();
                clusterId = clusterMapper.selectByUserId(parentUserId).getId();
            }
        } else {
            throw new RuntimeException("ProjectUtil--getRedisPrefixByUserIdAndProjectIdAndTaksId 未知账户类型，无法获取集群信息！");
        }
        String clusterPrefix = "cluster:" + clusterId;
        String clusterRunningPrefix = clusterPrefix + ":running";
        String clusterWaitingPrefix = clusterPrefix + ":waiting";
        String projectRunningKey = clusterRunningPrefix + ":" + projectId;
        String projectWaitingKey = clusterWaitingPrefix + ":" + projectId;
        String projectCheckKey = projectRunningKey + ":check";

        return PrefixTO.builder()
                .clusterPrefix(clusterPrefix)
                .clusterRunningPrefix(clusterRunningPrefix)
                .clusterWaitingPrefix(clusterWaitingPrefix)
                .projectRunningKey(projectRunningKey)
                .projectWaitingKey(projectWaitingKey)
                .projectCheckKey(projectCheckKey)
                .build();
    }


    public PrefixTO getRedisPrefixByClusterIdAndProjectId(String clusterId, String projectId) {
        String clusterPrefix = "cluster:" + clusterId;
        String clusterRunningPrefix = clusterPrefix + ":running";
        String clusterWaitingPrefix = clusterPrefix + ":waiting";
        String projectRunningKey = clusterRunningPrefix + ":" + projectId;
        String projectWaitingKey = clusterWaitingPrefix + ":" + projectId;
        String projectCheckKey = clusterWaitingPrefix + ":" + projectId + ":check";

        return PrefixTO.builder()
                .clusterPrefix(clusterPrefix)
                .clusterRunningPrefix(clusterRunningPrefix)
                .clusterWaitingPrefix(clusterWaitingPrefix)
                .projectRunningKey(projectRunningKey)
                .projectWaitingKey(projectWaitingKey)
                .projectCheckKey(projectCheckKey)
                .build();

    }


    /**
     * 获取 projectId 列表
     *
     * @param clusterRunningKeySet 集群下的所有键值对（包括运行中的项目和等待中的项目）
     * @return projectId 列表
     */
    public List<String> getRunningProjectList(Set<String> clusterRunningKeySet) {
        return clusterRunningKeySet.stream().filter(key -> StringUtil.countSubString(key, ":") == 3).collect(Collectors.toList());
    }

    public void sendMessage(String topic, String message) {
        kafkaTemplate.send(topic, message);
    }


    public PrefixTO getRedisPrefixByProjectIdAndProjectType(String projectId, String projectType) {
        String userId;
        if (DictConstants.PROJECT_TYPE_MANUAL.equals(projectType)) {
            userId = manualProjectMapper.selectCreateUserById(projectId);
        } else if (DictConstants.PROJECT_TYPE_AUTO_SUB.equals(projectType)) {
            userId = autoSubProjectMapper.selectCreateUserById(projectId);
        } else {
            throw new RuntimeException("PrjectUtil--getRedisPrefixByProjectIdAndProjectType 未知的项目类型！");
        }

        //3 获取用户类型（管理员账户、管理员子账户、普通账户、普通子账户）（独占、共享）
        UserPO userPO = userMapper.selectById(userId);
        String roleCode = userPO.getRoleCode();
        String useType = userPO.getUseType();
        String clusterId;
        if (DictConstants.ROLE_CODE_SYSADMIN.equals(roleCode) || DictConstants.ROLE_CODE_ADMIN.equals(roleCode)) {  //3-1 管理员账户和管理员子账户直接执行
            clusterId = DictConstants.SYSTEM_CLUSTER_ID;
        } else if (DictConstants.ROLE_CODE_UESR.equals(roleCode)) { //3-2 普通账户，不管是独占还是共享，都在自己的集群里排队，根据自己的独占节点排队
            clusterId = clusterMapper.selectByUserId(userId).getId();
        } else if (DictConstants.ROLE_CODE_SUBUESR.equals(roleCode)) {
            if (DictConstants.USER_TYPE_EXCLUSIVE.equals(useType)) {   //3-3 普通子账户，根据自己的独占节点排队
                clusterId = clusterMapper.selectByUserId(userId).getId();
            } else {    //3-4 共享子账户，根据父账户的共享节点排队
                String parentUserId = userPO.getCreateUserId();
                clusterId = clusterMapper.selectByUserId(parentUserId).getId();
            }
        } else {
            throw new RuntimeException("ProjectUtil--getRedisPrefixByUserIdAndProjectIdAndTaksId 未知账户类型，无法获取集群信息！");
        }
        String clusterPrefix = "cluster:" + clusterId;
        String clusterRunningPrefix = clusterPrefix + ":running";
        String clusterWaitingPrefix = clusterPrefix + ":waiting";
        String projectRunningKey = clusterRunningPrefix + ":" + projectId;
        String projectWaitingKey = clusterWaitingPrefix + ":" + projectId;
        String projectCheckKey = projectRunningKey + ":check";

        return PrefixTO.builder()
                .clusterPrefix(clusterPrefix)
                .clusterRunningPrefix(clusterRunningPrefix)
                .clusterWaitingPrefix(clusterWaitingPrefix)
                .projectRunningKey(projectRunningKey)
                .projectWaitingKey(projectWaitingKey)
                .projectCheckKey(projectCheckKey)
                .build();
    }

    /**
     * 将 redis 中的变量 +1，需要保证同步
     *
     * @param nodeName 节点名称
     */
    @Synchronized
    public void addOneParallelismToNode(String nodeName) {
        String key = "node:" + nodeName + ":parallelism";
        String parallelismString = stringRedisTemplate.opsForValue().get(key);
        if (StringUtil.isEmpty(parallelismString)) {
            throw new RuntimeException("ProjectUtil--addOneParallelismToNode redisKey " + key + " 为空。");
        }
        int parallelism = Integer.parseInt(parallelismString);
        stringRedisTemplate.opsForValue().set(key, (parallelism + 1) + "");
        log.info("ProjectUtil--addOneParallelismToNode 归还节点 " + nodeName + " 并行度从 " + parallelism + " 加一。");
    }
}
